import re, os.path
from SCons.Options import Options
from SCons.Environment import Environment

def pathValidator(key, val, env):
    """ Silence SCons with regard to non-existent paths """
    pass

def PathOption(dir, explanation, default):
    """ Define PathOption that works the same across old and new (0.96.9*) SCons, i.e. new SCons
    adds a restrictive path validator by default. 
    Used when setting commandline options in the Configure object.
    """
    return (dir, "%s ( /path/to/%s )" % (explanation, dir), default, pathValidator)

class Configure(object):
    """ Replace the standard SCons Configure method.
    
    The standard SCons Configure method is somewhat weak, in particular with
    regard to cleaning up after previous builds. When cleaning you normally
    don't want to reconfigure, you just want to clean the targets from the
    previous build. You may still want to get at configuration parameters
    however, since they may affect which targets are built (e.g. a module is
    only built in the presence of an external library). This class caches
    configuration values between builds, even command-line options, so that the
    build configuration when cleaning mirrors that when building.
    """
    def __init__(self, env, arguments, options=[], customTests={}):
        """ @param env: SCons environment
        @param arguments: SCons command-line arguments (ARGUMENTS)
        @param options: Command-line options.
        """
        useCache = env.GetOption("clean") # Prefer cached values if cleaning
        
        self._conf = env.Configure(log_file=os.path.join("scons", "configure.log"), custom_tests=\
                customTests)
        for n, f in customTests.items():
            setattr(self, n, getattr(self._conf, n))
        self._oldCache, self._newCache = {}, {}
        self._cachePath = os.path.abspath(os.path.join("scons", "configure.cache"))
        if useCache and os.path.exists(self._cachePath):
            f = file(self._cachePath)
            try:
                context = None
                import re
                for l in f:
                    m = re.match(r"(.+):", l)
                    if m:
                        context = m.group(1)
                        self._oldCache[context] = {}
                        continue
                    if context is None:
                        continue

                    m = re.match("(.+) = (.+)", l)
                    try:
                        k, v = m.groups()
                    except:
                        continue
                    self._oldCache[context][k] = v
            finally:
                f.close()

        args = arguments.copy()
        optsCache = os.path.join("scons", "options.cache")   # Save options between runs in this cache
        if useCache:
            # Then ignore new values
            for o in options:
                if o[0] in args:
                    del args[o[0]]

        opts = Options(optsCache, args=args)
        opts.AddOptions(*options)
        opts.Update(env)
        self.__opts = []
        for o in options:
            self.__opts.append(env[o[0]])
        # Cache options if asked to
        cacheOptions = env.get("cacheOptions", False)
        if cacheOptions:
            del env["cacheOptions"] # Don't store this value
            opts.Save(optsCache, env)
            # Restore
            env["cacheOptions"] = cacheOptions
        env.Help(opts.GenerateHelpText(env))
    
    def checkCxxHeader(self, header, include_quotes='""'):
        """ Look for C++ header.

        @param header: The header to look for. This may be a sequence, in which case the preceding
        items are headers to be included before the required header.
        """
        ctxt = "checkCxxHeader"
        try:
            r = self._oldCache[ctxt][str(header)]
        except KeyError:
            r = self._conf.CheckCXXHeader(header, include_quotes)

        if not ctxt in self._newCache:
            self._newCache[ctxt] = {}
        self._newCache[ctxt][str(header)] = r
        return r

    def checkLibWithHeader(self, lib, header, language="C", call=None):
        ctxt = "checkLibWithHeader"
        try:
            r = self._oldCache[ctxt][str(lib)]
        except KeyError:
            kwds = {}
            if call is not None:
                kwds["call"] = call
            r = self._conf.CheckLibWithHeader(lib, header, language, **kwds)

        if not ctxt in self._newCache:
            self._newCache[ctxt] = {}
        self._newCache[ctxt][str(lib)] = r
        return r

    def finish(self):
        """ Update configuration cache """
        f = file(self._cachePath, "w")
        try:
            for ctxt, d in self._newCache.items():
                f.write("%s:\n" % ctxt)
                for k, v in d.items():
                    f.write("%s = %s\n" % (k, v))
        finally:
            f.close()

        return self.__opts
